// Copyright 2012, 2013 Brad Block, Pawjaw, LLC. (an Ohio Limited Liability Company)
// 
// This file is part of JBTCRF.
// 
// JBTCRF is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// JBTCRF is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with JBTCRF.  If not, see <http://www.gnu.org/licenses/>.

package com.pawjaw.classification.crf.lmcbt;

import com.pawjaw.classification.crf.lmcbt.configurations.Configuration;
import com.pawjaw.classification.crf.lmcbt.points.Point;
import java.util.List;
import com.pawjaw.classification.crf.lmcbt.trees.BoostedTrees;
import com.pawjaw.classification.crf.lmcbt.sequences.ExpandedPointSequence;

import static com.pawjaw.classification.crf.lmcbt.CRFMath.*;

/**
 * Class to take trained and serialized CRF model and apply to new data for
 * labeling.  Create new instance then use set() method to specify model and
 * configuration.  Configuration specified should match that used to train
 * the specified CRF model.
 */
public class CRFInference {
    private Configuration c;
    private double[][][] sbt_k;

    /**
     * Set the configuration options and the CRF model.  Configuration options
     * should match those used to train CRF model.
     *
     * @param c configuration options
     * @param sbt_k serialized CRF model
     */
    public void set(Configuration c, double[][][] sbt_k) {
        this.c = c;
        this.sbt_k = sbt_k;
    }

    private double potential(int label, int sequence_position, int previous_label, ExpandedPointSequence eps) {
        return BoostedTrees.boostedOutput(c, eps, sequence_position, previous_label, sbt_k[label]);
    }

    /**
     * Label given sequence and compare to given labels for testing purposes.
     * Use Viterbi method for labeling (most likely overall label sequence).
     *
     * @param eps sequence to label
     * @param true_labels true labels of the points in the sequence
     * @return number of labels correctly identified by the CRF model
     */
    public int correctVirterbi(ExpandedPointSequence eps, int[] true_labels) {
        int i = 0, correct = 0;
        for(int found_label : labelViterbi(eps))
            if(found_label == true_labels[i++])
                correct++;
        return correct;
    }

    /**
     * Label given sequence and compare to given labels for testing purposes.
     * Use Forward-Backward method for labeling (most likely individual
     * point-wise labels).
     *
     * @param eps sequence to label
     * @param true_labels true labels of the points in the sequence
     * @return number of labels correctly identified by the CRF model
     */
    public int correctForwardBackward(ExpandedPointSequence eps, int[] true_labels) {
        int i = 0, correct = 0;
        for(int found_label : labelForwardBackward(eps))
            if(found_label == true_labels[i++])
                correct++;
        return correct;
    }

    /**
     * Label given sequence sand compare to given labels for testing purposes.
     * Use Viterbi method for labeling (most likely overall label sequence).
     *
     * @param pointss sequences to label
     * @param true_labelss true labels of the points in the sequences
     * @return percentage of correctly labeled points
     */
    public double accuracyViterbi(List<Point[]> pointss, List<int[]> true_labelss) {
        int i = 0, correct = 0, total = 0;
        ExpandedPointSequence eps = new ExpandedPointSequence();
        for(Point[] points : pointss) {
            eps.set(points);
            correct += correctVirterbi(eps, true_labelss.get(i++));
            total += points.length;
        }
        return (double)correct / (double)total;
    }

    /**
     * Label given sequence sand compare to given labels for testing purposes.
     * Use Forward-Backward method for labeling (most likely individual
     * point-wise labels).
     *
     * @param pointss sequences to label
     * @param true_labelss true labels of the points in the sequences
     * @return percentage of correctly labeled points
     */
    public double accuracyForwardBackward(List<Point[]> pointss, List<int[]> true_labelss) {
        int i = 0, correct = 0, total = 0;
        ExpandedPointSequence eps = new ExpandedPointSequence();
        for(Point[] points : pointss) {
            eps.set(points);
            correct += correctForwardBackward(eps, true_labelss.get(i++));
            total += points.length;
        }
        return (double)correct / (double)total;
    }

    /**
     * Infer the labels of the given sequence using the Viterbi method
     * (most likely overall label sequence).  Labels are in the range specified
     * in the configuration given to the set() method.
     *
     * @param eps given sequence to label
     * @return the labels of each point in the sequence
     */
    public int[] labelViterbi(ExpandedPointSequence eps) {
        int t, T = eps.length(), k, kp, maxk, K = sbt_k.length - 1, maxarg;
        double logtransition, logmaxtransition;
        int[] label_t = new int[T];
        int[][] maxarg_k_t = new int[T][K + 1];
        double[][] logmaxtransition_k_t = new double[T][K + 1];
        for(k = 0;k <= K;k++) {
            maxarg_k_t[0][k] = -1;
            logmaxtransition_k_t[0][k] = LOGONE;
        }
        for(t = 1;t < T;t++)
            for(k = 0;k <= K;k++) {
                maxarg = -1;
                logmaxtransition = Double.NEGATIVE_INFINITY;
                for(kp = 0;kp <= K;kp++)
                    if((logtransition = eLogProduct(potential(k, t, kp, eps), logmaxtransition_k_t[t - 1][kp])) > logmaxtransition) {
                        maxarg = kp;
                        logmaxtransition = logtransition;
                    }
                maxarg_k_t[t][k] = maxarg;
                logmaxtransition_k_t[t][k] = logmaxtransition;
            }
        maxarg = -1;
        logmaxtransition = Double.NEGATIVE_INFINITY;
        for(k = 0;k < K;k++)
            if(logmaxtransition_k_t[T - 1][k] > logmaxtransition) {
                maxarg = k;
                logmaxtransition = logmaxtransition_k_t[T - 1][k];
            }
        label_t[T - 1] = maxarg;
        for(t = T - 2;t >= 0;t--)
            label_t[t] = maxarg_k_t[t + 1][label_t[t + 1]];
        return label_t;
    }

    /**
     * Infer the labels of the given sequence using the Forward-Backward method
     * (most likely individual point-wise labels ).  Labels are in the range
     * specified in the configuration given to the set() method.
     *
     * @param eps given sequence to label
     * @return the labels of each point in the sequence
     */
    public int[] labelForwardBackward(ExpandedPointSequence eps) {
        int t, k, kp, T = eps.length(), K = sbt_k.length - 1, maxarg;
        double logsum, logfb, maxlogfb;
        int[] label_t = new int[T];
        double[][] logalpha_k_ta = new double[T + 1][K + 1];
        double[][] logbeta_k_ta = new double[T + 1][K + 1];
        for(k = 0;k <= K;k++)
            logalpha_k_ta[1][k] = potential(k, 0, K, eps);
        for(t = 1;t < T;t++) {
            for(k = 0;k <= K;k++) {
                logsum = LOGZERO;
                for(kp = 0;kp <= K;kp++)
                    logsum = eLogSum(logsum, eLogProduct(potential(k, t, kp, eps), logalpha_k_ta[t][kp]));
                logalpha_k_ta[t + 1][k] = logsum;
            }
        }
        for(t = T - 1;t >= 0;t--) {
            for(k = 0;k <= K;k++) {
                logsum = LOGZERO;
                for(kp = 0;kp <= K;kp++)
                    logsum = eLogSum(logsum, eLogProduct(potential(k, t, kp, eps), logbeta_k_ta[t + 1][kp]));
                logbeta_k_ta[t][k] = logsum;
            }
        }
        for(t = 0;t < T;t++) {
            maxarg = -1;
            maxlogfb = Double.NEGATIVE_INFINITY;
            for(k = 0;k < K;k++)
                if((logfb = eLogProduct(logalpha_k_ta[t + 1][k], logbeta_k_ta[t + 1][k])) > maxlogfb) {
                    maxarg = k;
                    maxlogfb = logfb;
                }
            label_t[t] = maxarg;
        }
        return label_t;
    }
}
