// Copyright 2012, 2013 Brad Block, Pawjaw, LLC. (an Ohio Limited Liability Company)
// 
// This file is part of JBTCRF.
// 
// JBTCRF is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// JBTCRF is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with JBTCRF.  If not, see <http://www.gnu.org/licenses/>.

package com.pawjaw.classification.crf.lmcbt;

import com.pawjaw.classification.crf.lmcbt.configurations.Configuration;
import com.pawjaw.classification.crf.lmcbt.trees.BoostedTrees;
import com.pawjaw.classification.crf.lmcbt.sequences.TrainingSequences;
import com.pawjaw.classification.crf.lmcbt.points.Point;
import java.util.List;
import java.util.concurrent.Semaphore;

/**
 * Main entry point for training new CRF model.  Following instantiation,
 * set all training examples using set() method, then run one of the
 * train() methods.
 * 
 */
public class CRFTrainer {
    private Configuration c;
    private TrainingSequences ts;
    private BoostedTrees[] bt_k;

    /**
     * Instantiate this, then use set() repeatedly finally followed by train().
     * Same configuration should be used for inference.
     *
     * @param c options for training specified here
     * @param example_sequences number of example sequences to be used in training
     */
    public CRFTrainer(Configuration c, int example_sequences) {
        this.c = c;
        ts = new TrainingSequences();
        ts.allocate(c, example_sequences);
        bt_k = new BoostedTrees[c.label_count_including_start_label];
    }

    /**
     * Set a new training example.  sequence_index ranges up to example_sequences
     * specified in constructor.  True labels range to the value specified in
     * the configuraton given to the constructor.
     *
     * @param sequence_index ranges from 0 to example_sequences - 1 (inclusive)
     * @param points data points in the given sequence
     * @param labels true labels for the given sequence
     */
    public void set(int sequence_index, Point[] points, int[] labels) {
        ts.set(sequence_index, points, labels);
    }

    /**
     * Train a new CRF model with the number of boosting iterations specified
     * in the configuration given to the constructor.
     */
    public void train() {
        int k, K = bt_k.length, m;
        Semaphore update_complete = new Semaphore(0);
        Semaphore proceed = new Semaphore(c.boosted_tree_threads);
        ts.indexExamples();
        if(c.cache_expanded_true_features)
            ts.cacheTrueFeatures();
        for(k = 0;k < K;k++)
            (bt_k[k] = new BoostedTrees(c, ts, k, proceed, update_complete)).start();
        for(m = 1;m <= c.max_boosting_iterations;m++) {
            if(c.report_training_progress)
                System.err.println("boosting iteration: " + m + " / " + c.max_boosting_iterations);
            ts.update();
            for(k = 0;k < K;k++)
                bt_k[k].setUpdateable();
            try {
                update_complete.acquire(K);
            } catch(InterruptedException e) {
                break;
            }
        }
        for(k = 0;k < K;k++)
            bt_k[k].finished();
    }

    /**
     * Train a new CRF model with the maximum number of boosting iterations
     * specified in the configuration given to the constructor but with the
     * option to stop early if the incremental improvement in Viterbi accuracy
     * is less than the amount specified in the configuration.  Use the
     * arguments to this method to validate the model for accuracy.
     *
     * @param pointss validation data points
     * @param true_labelss validation labels
     */
    public void train(List<Point[]> pointss, List<int[]> true_labelss) {
        int i = c.boosting_iterations_between_tests, k, K = bt_k.length, m;
        double accuracy, previous_accuracy = 0;
        CRFInference crfi = new CRFInference();
        Semaphore update_complete = new Semaphore(0);
        Semaphore proceed = new Semaphore(c.boosted_tree_threads);
        ts.indexExamples();
        if(c.cache_expanded_true_features)
            ts.cacheTrueFeatures();
        for(k = 0;k < K;k++)
            (bt_k[k] = new BoostedTrees(c, ts, k, proceed, update_complete)).start();
        for(m = 1;m <= c.max_boosting_iterations;m++) {
            if(c.report_training_progress)
                System.err.println("boosting iteration: " + m + " / " + c.max_boosting_iterations);
            if(m >= c.min_boosting_iterations && i++ == c.boosting_iterations_between_tests) {
                crfi.set(c, serialize());
                accuracy = crfi.accuracyViterbi(pointss, true_labelss);
                if(c.report_training_progress)
                    System.err.println("viterbi validation accuracy at iteration " + m + ": " + accuracy);
                if(previous_accuracy > 0 && (accuracy - previous_accuracy) / previous_accuracy < c.min_relative_accuracy_improvement) {
                    System.err.println("stopping early, relative accuracy gained: " + ((accuracy - previous_accuracy) / previous_accuracy));
                    break;
                } else {
                    i = 1;
                    previous_accuracy = accuracy;
                }
            }
            ts.update();
            for(k = 0;k < K;k++)
                bt_k[k].setUpdateable();
            try {
                update_complete.acquire(K);
            } catch(InterruptedException e) {
                break;
            }
        }
        for(k = 0;k < K;k++)
            bt_k[k].finished();
    }

    /**
     * Serialize trained CRF model to be written to disk and/or used for
     * subsequent inference (labeling new data).
     *
     * @return CRF model for subsequent inference
     */
    public double[][][] serialize() {
        int k, K = bt_k.length;
        double[][][] sbt_k = new double[K][][];
        for(k = 0;k < K;k++)
            sbt_k[k] = bt_k[k].serialize();
        return sbt_k;
    }
}
