// Copyright 2012, 2013 Brad Block, Pawjaw, LLC. (an Ohio Limited Liability Company)
// 
// This file is part of JBTCRF.
// 
// JBTCRF is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// JBTCRF is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with JBTCRF.  If not, see <http://www.gnu.org/licenses/>.

package com.pawjaw.classification.crf.lmcbt.configurations;

/**
 * Single point for configuring all options for training and inference.
 */
public abstract class Configuration {
    /**
     * One side of the sliding window (e.g. window_radius = 3 means
     * use features from previous 3 and next 3 elements in the
     * sequence).
     */
    public final int window_radius;
    /**
     * Number of features in each individual element (aka point).
     */
    public final int point_features;
    /**
     * Number of threads used in multithreaded sufficient
     * statistics collection portion of node splitting during
     * regression tree construction.
     */
    public final int splitter_threads = getSplitterThreads();
    /**
     * Number of features in each individual element (aka point)
     * including features incorporated through sliding window and
     * previous point label.
     */
    public final int expanded_features;
    /**
     * Offset into feature index where valid window features start.
     */
    public final int valid_window_offset;
    /**
     * Maximum leaves per boosted regression tree.
     */
    public final int max_leaves_per_tree = getMaxLeavesPerTree();
    /**
     * Number of threads used in multithreaded tree boosting.
     * Can utilize as many simultaneous threads as the value in
     * {@link label_count_including_start_label}.
     */
    public final int boosted_tree_threads = getBoostedTreeThreads();
    /**
     * Maximum number of boosted tree iterations.  May be less if
     * training includes validation data and
     * {@link min_relative_accuracy_improvement} is greater than zero.
     */
    public final int max_boosting_iterations = getMaxBoostingIterations();
    /**
     * Minimum number of boosting iterations regardless of validation
     * data and {@link min_relative_accuracy_improvement}.
     */
    public final int min_boosting_iterations = getMinBoostingIterations();
    /**
     * Boosting itereations to run between evaluating stopping criteria when
     * validation data is provided for training.
     */
    public final int boosting_iterations_between_tests = getBoostingIterationsBetweenTests();
    /**
     * Number of labels for elements in the data not including the start
     * pseudo-label.
     */
    public final int label_count_excluding_start_label;
    /**
     * Number of labels for elements in the data including the start
     * pseudo-label.
     */
    public final int label_count_including_start_label;
    /**
     * Shrinkage for regression tree smoothing.  Typical range is 0 to 100.
     */
    public final double regression_tree_shrinkage = getRegressionTreeShrinkage();
    /**
     * Minimum relative accuracy improvement required to continue boosting
     * when training includes validation data.  Typical range is 1e-3 to 1e-4.
     */
    public final double min_relative_accuracy_improvement = getMinRelativeAccuracyImprovement();
    /*
     * Print to standard error some basic reporting of training progress.
     */
    public final boolean report_training_progress = getReportTrainingProgress();
    /*
     * Cache expanded true features for training examples.
     */
    public final boolean cache_expanded_true_features = getCacheExpandedTrueFeatures();

    protected abstract int getSplitterThreads();
    protected abstract int getMaxLeavesPerTree();
    protected abstract int getBoostedTreeThreads();
    protected abstract int getMaxBoostingIterations();
    protected abstract int getMinBoostingIterations();
    protected abstract int getBoostingIterationsBetweenTests();
    protected abstract double getRegressionTreeShrinkage();
    protected abstract double getMinRelativeAccuracyImprovement();
    protected abstract boolean getReportTrainingProgress();
    protected abstract boolean getCacheExpandedTrueFeatures();

    public Configuration(int point_features, int window_radius, int label_count_excluding_start_label) {
        this.point_features = point_features;
        this.window_radius = window_radius;
        this.label_count_excluding_start_label = label_count_excluding_start_label;
        label_count_including_start_label = label_count_excluding_start_label + 1;
        valid_window_offset = label_count_including_start_label + point_features * (window_radius * 2 + 1);
        expanded_features = valid_window_offset + window_radius * 2 + 1;
    }
}
