// Copyright 2012, 2013 Brad Block, Pawjaw, LLC. (an Ohio Limited Liability Company)
// 
// This file is part of JBTCRF.
// 
// JBTCRF is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// JBTCRF is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with JBTCRF.  If not, see <http://www.gnu.org/licenses/>.

package com.pawjaw.classification.crf.lmcbt.sequences;

import com.pawjaw.classification.crf.lmcbt.configurations.Configuration;

import static com.pawjaw.classification.crf.lmcbt.CRFMath.*;

public class TrainingSequence extends ExpandedPointSequence {
    private int[] label_t; // T length vector of labels in range (0 to K - 1 inclusive)
    private double[][] logalpha_k_ta; // T + 1 x K + 1 forward recursion matrix (ta index is augmented with start state)
    private double[][] logbeta_k_ta; //  T + 1 x K + 1 backward recursion matrix (ta index is augmented with start state)
    private double[][][] target_kp_k_t; //  T square matrices length K + 1 for holding regression targets
    private double[][][] potential_kp_k_t; //  T square matrices length K + 1 for holding potential values

    public void allocate(Configuration c) {
        int T = length(), K = c.label_count_excluding_start_label;
        logalpha_k_ta = new double[T + 1][K + 1];
        logbeta_k_ta = new double[T + 1][K + 1];
        target_kp_k_t = new double[T][K + 1][K + 1];
        potential_kp_k_t = new double[T][K + 1][K + 1];
        initializePotential(T, K);
        initializeForwardBackward(T, K);
    }

    public void setLabels(int[] label_t) {
        this.label_t = label_t;
    }

    // k = K is start label position
    private void initializeForwardBackward(int T, int K) {
        int k;
        for(k = 0;k < K;k++)
            logalpha_k_ta[0][k] = LOGZERO;
        logalpha_k_ta[0][K] = LOGONE;
        for(k = 0;k <= K;k++)
            logbeta_k_ta[T][k] = LOGONE;
    }

    private void initializePotential(int T, int K) {
        int t, k, kp;
        for(t = 0;t < T;t++)
            for(k = 0;k <= K;k++)
                for(kp = 0;kp <= K;kp++)
                    potential_kp_k_t[t][k][kp] = 0;
    }

    public void updateForwardBackward() {
        int t, k, kp, T = length(), K = logalpha_k_ta[0].length - 1;
        double logsum;
        for(k = 0;k <= K;k++)
            logalpha_k_ta[1][k] = potential_kp_k_t[0][k][K];
        for(t = 1;t < T;t++) {
            for(k = 0;k <= K;k++) {
                logsum = LOGZERO;
                for(kp = 0;kp <= K;kp++)
                    logsum = eLogSum(logsum, eLogProduct(potential_kp_k_t[t][k][kp], logalpha_k_ta[t][kp]));
                logalpha_k_ta[t + 1][k] = logsum;
            }
        }
        for(t = T - 1;t >= 0;t--) {
            for(k = 0;k <= K;k++) {
                logsum = LOGZERO;
                for(kp = 0;kp <= K;kp++)
                    logsum = eLogSum(logsum, eLogProduct(potential_kp_k_t[t][k][kp], logbeta_k_ta[t + 1][kp]));
                logbeta_k_ta[t][k] = logsum;
            }
        }
    }
    
    public void updateTargets() {
        int t, k, kp, T = length(), K = logalpha_k_ta[0].length - 1;
        double lognormalizer = 0, logtarget;
        for(t = 0;t < T;t++) {
            lognormalizer = LOGZERO;
            for(k = 0;k <= K;k++)
                for(kp = 0;kp <= K;kp++) {
                    logtarget = eLogProduct(logalpha_k_ta[t][kp],
                            eLogProduct(potential_kp_k_t[t][k][kp], logbeta_k_ta[t + 1][k]));
                    target_kp_k_t[t][k][kp] = logtarget;
                    lognormalizer = eLogSum(lognormalizer, logtarget);
                }
            for(k = 0;k <= K;k++)
                for(kp = 0;kp <= K;kp++)
                    target_kp_k_t[t][k][kp] = eExp(eLogProduct(target_kp_k_t[t][k][kp], -lognormalizer));
        }
        for(t = 0;t < T;t++)
            for(k = 0;k <= K;k++)
                for(kp = 0;kp <= K;kp++)
                    if(label_t[t] == k) {
                        if(t == 0) {
                            if(kp == K)
                                target_kp_k_t[t][k][kp] = 1.0 - target_kp_k_t[t][k][kp];
                            else
                                target_kp_k_t[t][k][kp] = 0.0 - target_kp_k_t[t][k][kp];
                        } else {
                            if(kp == label_t[t - 1])
                                target_kp_k_t[t][k][kp] = 1.0 - target_kp_k_t[t][k][kp];
                            else
                                target_kp_k_t[t][k][kp] = 0.0 - target_kp_k_t[t][k][kp];
                        }
                    } else
                        target_kp_k_t[t][k][kp] = 0.0 - target_kp_k_t[t][k][kp];
    }

    public double getTarget(int label, int example_index) {
        int K = logalpha_k_ta[0].length - 1;
        int t = example_index / (K + 1);
        int kp = example_index % (K + 1);
        return target_kp_k_t[t][label][kp];
    }

    public void incrementPotential(int label, int example_index, double potential) {
        int K = logalpha_k_ta[0].length - 1;
        int t = example_index / (K + 1);
        int kp = example_index % (K + 1);
        potential_kp_k_t[t][label][kp] += potential;
    }
}
