// Copyright 2012, 2013 Brad Block, Pawjaw, LLC. (an Ohio Limited Liability Company)
// 
// This file is part of JBTCRF.
// 
// JBTCRF is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// JBTCRF is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with JBTCRF.  If not, see <http://www.gnu.org/licenses/>.

package com.pawjaw.classification.crf.lmcbt.trees;

import com.pawjaw.classification.crf.lmcbt.configurations.Configuration;
import com.pawjaw.classification.crf.lmcbt.sequences.TrainingSequences;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.concurrent.Semaphore;

public class Splitter {
    private int features, count, label;
    private double sum;
    private Configuration c;
    private TrainingSequences ts;
    private Semaphore proceed, update_complete;
    private int[] count_f, example_indexes;
    private double[] sum_f;
    private StatsThread[] sts;

    public Splitter(Configuration c, TrainingSequences ts) {
        int i;
        this.c = c;
        this.ts = ts;
        features = ts.features();
        count_f = new int[features];
        sum_f = new double[features];
        update_complete = new Semaphore(0);
        sts = new StatsThread[c.splitter_threads];
        proceed = new Semaphore(c.splitter_threads);
        for(i = 0;i < c.splitter_threads;i++)
            (sts[i] = new StatsThread(i)).start();
    }

    public void initialize(int label, Node root) {
        double sum = 0;
        this.label = label;
        if(root.example_indexes.length == 0)
            root.output = 0;
        else {
            for(int example_index : root.example_indexes)
                sum += ts.getTarget(label, example_index);
            root.output = sum / (c.regression_tree_shrinkage + root.example_indexes.length);
        }
    }

    private class StatsThread extends Thread {
        private int id;
        private Semaphore updateable = new Semaphore(0);
        private List<Integer> feature_indexes = new ArrayList();

        public double _sum;
        public int[] _count_f = new int[features];
        public double[] _sum_f = new double[features];

        public StatsThread(int id) {
            this.id = id;
        }

        public void run() {
            int i, I, length;
            double target;
            try {
                while(true) {
                    updateable.acquire();
                    proceed.acquire();
                    _sum = 0;
                    Arrays.fill(_sum_f, 0);
                    Arrays.fill(_count_f, 0);
                    length = (int)Math.ceil((double)example_indexes.length / (double)sts.length);
                    I = Math.min(example_indexes.length, length * (id + 1));
                    for(i = length * id;i < I;i++) {
                        _sum += target = ts.getTarget(label, example_indexes[i]);
                        if(c.cache_expanded_true_features)
                            for(int feature_index : ts.getCachedTrueFeatures(example_indexes[i])) {
                                _count_f[feature_index]++;
                                _sum_f[feature_index] += target;
                            }
                        else
                            for(int feature_index : ts.getTrueFeatures(example_indexes[i], feature_indexes)) {
                                _count_f[feature_index]++;
                                _sum_f[feature_index] += target;
                            }
                    }
                    proceed.release();
                    update_complete.release();
                }
            } catch(InterruptedException e) {}
        }

        public void setUpdateable() {
            updateable.release();
        }

        public void finished() {
            this.interrupt();
        }
    }

    public void finished() {
        int i;
        for(i = 0;i < sts.length;i++)
            sts[i].finished();
    }

    private void collectStatistics(int[] example_indexes) {
        int i, j;
        this.example_indexes = example_indexes;
        for(i = 0;i < sts.length;i++)
            sts[i].setUpdateable();
        Arrays.fill(sum_f, 0);
        Arrays.fill(count_f, 0);
        try {
            update_complete.acquire(sts.length);
        } catch(InterruptedException e) {
            throw new RuntimeException(e);
        }
        sum = 0;
        count = example_indexes.length;
        for(i = 0;i < features;i++) {
            for(j = 0;j < sts.length;j++) {
                sum_f[i] += sts[j]._sum_f[i];
                count_f[i] += sts[j]._count_f[i];
            }
        }
        for(j = 0;j < sts.length;j++)
            sum += sts[j]._sum;
    }

    public Split findBestSplit(Node parent) {
        int f;
        double true_count, false_count, gain;
        Split s = new Split(parent);
        collectStatistics(parent.example_indexes);
        for(f = 0;f < features;f++)
            if((true_count = count_f[f]) >= 2 && (false_count = count - count_f[2]) >= 2) {
                s.true_child_output = sum_f[f] / (c.regression_tree_shrinkage + true_count);
                s.false_child_output = (sum - sum_f[f]) / (c.regression_tree_shrinkage + false_count);
                gain = (2.0 * c.regression_tree_shrinkage + true_count) * s.true_child_output * s.true_child_output +
                        (2.0 * c.regression_tree_shrinkage + false_count) * s.false_child_output * s.false_child_output -
                        (2.0 * c.regression_tree_shrinkage + count) * parent.output * parent.output;
                if(gain > s.gain) {
                    s.split_feature = f;
                    s.gain = gain;
                }
            }
        return s.isValid() ? s : null;
    }

    public Node[] acceptSplit(Split s) {
        List<Integer> true_child_example_indexes = new ArrayList();
        List<Integer> false_child_example_indexes = new ArrayList();
        for(int example_index : s.parent.example_indexes)
            if(ts.hasTrueFeature(example_index, s.split_feature))
                true_child_example_indexes.add(example_index);
            else
                false_child_example_indexes.add(example_index);
        s.parent.example_indexes = null;
        s.parent.split_feature = s.split_feature;
        return new Node[] {
            s.parent.true_child = new Node(s.true_child_output, true_child_example_indexes),
            s.parent.false_child = new Node(s.false_child_output, false_child_example_indexes)
        };
    }
}
