// Copyright 2012, 2013 Brad Block, Pawjaw, LLC. (an Ohio Limited Liability Company)
// 
// This file is part of JBTCRF.
// 
// JBTCRF is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// JBTCRF is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with JBTCRF.  If not, see <http://www.gnu.org/licenses/>.

package com.pawjaw.classification.crf.lmcbt.trees;

import com.pawjaw.classification.crf.lmcbt.configurations.Configuration;
import com.pawjaw.classification.crf.lmcbt.sequences.ExpandedPointSequence;
import com.pawjaw.classification.crf.lmcbt.sequences.TrainingSequences;
import java.util.ArrayList;
import java.util.LinkedList;
import java.util.List;
import java.util.PriorityQueue;
import java.util.Queue;

public class Tree {
    public Node root;

    private Queue<Split> splitq = new PriorityQueue();
    private List<Double> serializer_temp = new ArrayList();

    public void build(Configuration c, int label, Splitter splitter, List<Integer> example_indexes) {
        root = new Node(example_indexes);
        findAllSplits(c.max_leaves_per_tree, label, splitter);
    }

    public void build(Configuration c, int label, Splitter splitter, int[] example_indexes) {
        root = new Node(example_indexes);
        findAllSplits(c.max_leaves_per_tree, label, splitter);
    }

    public void printLeaves() {
        Node node;
        Queue<Node> nodeq = new LinkedList();
        nodeq.add(root);
        while(!nodeq.isEmpty())
            if((node = nodeq.poll()).isLeaf())
                System.out.println("    leaf output: " + node.output);
            else {
                nodeq.add(node.true_child);
                nodeq.add(node.false_child);
            }
    }

    public double mse(Configuration c, int label, TrainingSequences ts, List<Integer> example_indexes) {
        double mse = 0;
        Node node;
        for(int example_index : example_indexes) {
            node = root;
            while(!node.isLeaf())
                if(ts.hasTrueFeature(example_index, node.split_feature))
                    node = node.true_child;
                else
                    node = node.false_child;
            mse += Math.pow(node.output - ts.getTarget(label, example_index), 2.0);
        }
        return mse / (double)example_indexes.size();
    }

    public static double mse(Configuration c, int label, TrainingSequences ts, double[] st) {
        int example_index, examples = ts.examples();
        double mse = 0;
        for(example_index = 0;example_index < examples;example_index++)
            mse += Math.pow(Tree.output(ts, example_index, st) - ts.getTarget(label, example_index), 2.0);
        return mse / (double)examples;
    }

    private void findAllSplits(int max_leaves_per_tree, int label, Splitter splitter) {
        Split split;
        int leaves = 1;
        splitter.initialize(label, root);
        if((split = splitter.findBestSplit(root)) != null) {
            splitq.add(split);
            while(!splitq.isEmpty() && leaves < max_leaves_per_tree) {
                for(Node child : splitter.acceptSplit(splitq.poll()))
                    if((split = splitter.findBestSplit(child)) != null)
                        splitq.add(split);
                leaves++;
            }
        }
        splitq.clear();
    }

    public double[] serialize() {
        Node node;
        Integer child_pointer;
        Queue<Node> nodeq = new LinkedList();
        Queue<Integer> child_pointerq = new LinkedList();
        nodeq.add(root);
        child_pointerq.add(-1);
        while((node = nodeq.poll()) != null) {
            if((child_pointer = child_pointerq.poll()) >= 0)
                serializer_temp.set(child_pointer, (double)serializer_temp.size());
            if(node.isLeaf()) {
                serializer_temp.add(-1.0);
                serializer_temp.add(node.output);
            } else {
                serializer_temp.add((double)node.split_feature);
                nodeq.add(node.true_child);
                child_pointerq.add(serializer_temp.size());
                serializer_temp.add(-1.0);
                nodeq.add(node.false_child);
                child_pointerq.add(serializer_temp.size());
                serializer_temp.add(-1.0);
            }
        }
        return serializerTempToArray();
    }

    private double[] serializerTempToArray() {
        int i = 0;
        double[] serialized_tree = new double[serializer_temp.size()];
        for(double d : serializer_temp)
            serialized_tree[i++] = d;
        serializer_temp.clear();
        return serialized_tree;
    }

    public static double output(Configuration c, ExpandedPointSequence eps, int expanded_sequence_position, double[] serialized_tree) {
        return output(c, eps, expanded_sequence_position / c.label_count_including_start_label,
                expanded_sequence_position % c.label_count_including_start_label,
                serialized_tree);
    }

    public static double output(Configuration c, ExpandedPointSequence eps, int sequence_position, int previous_label, double[] serialized_tree) {
        int i = 0;
        while(i >= 0)
            if(serialized_tree[i] < 0)
                return serialized_tree[i + 1];
            else if(eps.hasTrueFeature(c, sequence_position, previous_label, (int)serialized_tree[i]))
                i = (int)serialized_tree[i + 1];
            else
                i = (int)serialized_tree[i + 2];
        return 0;
    }

    public static double output(TrainingSequences ts, int example_index, double[] serialized_tree) {
        int i = 0;
        while(i >= 0)
            if(serialized_tree[i] < 0)
                return serialized_tree[i + 1];
            else if(ts.hasTrueFeature(example_index, (int)serialized_tree[i]))
                i = (int)serialized_tree[i + 1];
            else
                i = (int)serialized_tree[i + 2];
        return 0;
    }

    public double output(Configuration c, ExpandedPointSequence eps, int sequence_position, int previous_label) {
        Node node = root;
        while(!node.isLeaf())
            if(eps.hasTrueFeature(c, sequence_position, previous_label, node.split_feature))
                node = node.true_child;
            else
                node = node.false_child;
        return node.output;
    }
}
